# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use
from .common import get_incident_tabs
import six
import logging

LOG = logging.getLogger(__name__)
TABS_LABEL = "step_label"

UI_TAB_ELEMENT = "tab"
UI_TAB_FIELD_TYPE = "incident"


class RequiredTabFields(type):
    """
    The job of this metaclass is to confirm that the user defined subclasses of Tab define
    required attributes, or otherwise throw an error at the time of class definition.
    """
    REQUIRED_TAB_ATTRIBUTES = ["UUID", "NAME", "SECTION", "CONTAINS"]

    def __init__(cls, name, bases, attrs):
        if not bases:
            return  # No bases implies Tab, which doesn't need to define the properties
        for attr in cls.REQUIRED_TAB_ATTRIBUTES:
            if not hasattr(cls, attr) or getattr(cls, attr) is None:
                raise AttributeError("{} is missing from class definition of a Tab".format(attr))


class Tab(six.with_metaclass(RequiredTabFields)):
    """
    Base class to be subclassed in integrations to create and update tabs related to those integrations.
    UUID, NAME, SECTION, and CONTAINS attributes have to be defined on the subclass.
    - UUID has to be a constant defined as a string to identify tab uniquely in all of the UIs.
    To generate UUID you can use python package `uuid` and method `uuid.uuid4()` in the python shell.
    - NAME is the name of the tab as it should appear in the UI
    - SECTION is integration's section in the app.config
    Can be used to place `ui_lock` in it to prevent changes being made to the tab.
    - CONTAINS - array of UI Elements (Field, Datatable) with api names to be listed in the tab

    Example usage:

    class QRadarTab(UITab):
        UUID = "abcdefg"
        NAME = "QRadar Tab"
        SECTION = "fn_qradar_integration"

        CONTAINS = [
            Field("api_name"),
            Datatable("api_name")
        ]

    create_tab(QRadarTab)
    """
    UUID = None
    NAME = None
    SECTION = None

    CONTAINS = None

    @classmethod
    def exists(cls, client):
        """
        Check if the tab exists in Resilient
        """
        return cls.exists_in(get_incident_tabs(client))

    @classmethod
    def exists_in(cls, tabs):
        """
        Checks if the tab exists in the list of tabs.
        """
        return cls.get_from_tabs(tabs) is not None

    @classmethod
    def get_from_tabs(cls, tabs):
        return next((x for x in tabs if x.get("predefined_uuid") == cls.UUID), None)

    @classmethod
    def as_dto(cls):
        if not hasattr(cls, "SHOW_IF"):
            cls.SHOW_IF = []
        return {
            "step_label": cls.NAME,
            "fields": [field.as_dto() for field in cls.CONTAINS] if cls.CONTAINS else [],
            "show_if": cls.SHOW_IF,
            "element": UI_TAB_ELEMENT,
            "field_type": UI_TAB_FIELD_TYPE,
            "predefined_uuid": cls.UUID,
            "show_link_header": False
        }

    @classmethod
    def get_missing_conditions(cls, tabs):
        if not hasattr(cls, "SHOW_IF"):
            return None
        if not cls.exists_in(tabs):
            return None
        required_conditions = cls.SHOW_IF
        tab = cls.get_from_tabs(tabs)
        present_conditions = tab.get("show_if", [])
        
        return [condition for condition in required_conditions if condition not in present_conditions]

    @classmethod
    def get_missing_fields(cls, tabs):
        """
        Given all the tabs find what fields are missing that are required in the `cls` tab.
        """
        if not cls.exists_in(tabs):
            return None
        tab = cls.get_from_tabs(tabs)
        tab_fields = tab.get('fields', [])
        return [field.as_dto() for field in cls.CONTAINS if not field.exists_in(tab_fields)]



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2019. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

import requests
import logging
from resilient_lib.components.integration_errors import IntegrationError
from resilient_lib.util.lib_common import deprecated


log = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 30


class RequestsCommon:
    """
    This class represents common functions around the use of the requests package for REST based APIs.
    It incorporates the app.config section "integrations" which can be used to define a common set of proxies
    for use by all functions using this library:

    [integrations]
    http_proxy=
    https_proxy=

    Similar properties may exist in the function's section which would override the [integrations] properties.
    """
    def __init__(self, opts=None, function_opts=None):
        # capture the properties for the integration as well as the global settings for all integrations for proxy urls
        self.integration_options = opts.get('integrations', None) if opts else None
        self.function_opts = function_opts

    def get_proxies(self):
        """ proxies can be specified globally for all integrations or specifically per function """
        proxies = None
        if self.integration_options and (self.integration_options.get("http_proxy") or self.integration_options.get("https_proxy")):
            proxies = {'http': self.integration_options.get("http_proxy"), 'https': self.integration_options.get("https_proxy")}

        if self.function_opts and (self.function_opts.get("http_proxy") or self.function_opts.get("https_proxy")):
            proxies = {'http': self.function_opts.get("http_proxy"), 'https': self.function_opts.get("https_proxy")}

        return proxies

    def get_timeout(self):
        """
        get the default timeout value in the 'integrations' section or the function's own section
        :return: timeout value
        """
        timeout = DEFAULT_TIMEOUT

        if self.integration_options and self.integration_options.get("timeout"):
            timeout = self.integration_options.get("timeout")

        if self.function_opts and self.function_opts.get("timeout"):
            timeout = self.function_opts.get("timeout")

        return int(timeout)

    def execute_call_v2(self, method, url, timeout=None, proxies=None, callback=None, **kwargs):
        """Constructs and sends a request. Returns :class:`Response` object.

            From the requests.requests() function, inputs are mapped to this function
            :param method: GET, HEAD, PATCH, POST, PUT, DELETE, OPTIONS
            :param url: URL for the request.
            :param params: (optional) Dictionary, list of tuples or bytes to send
                in the body of the :class:`Request`.
            :param data: (optional) Dictionary, list of tuples, bytes, or file-like
                object to send in the body of the :class:`Request`.
            :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
            :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
            :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
            :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
                ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
                or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content-type'`` is a string
                defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
                to add for the file.
            :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
            :param timeout: (optional) How many seconds to wait for the server to send data
                before giving up, as a float, or a :ref:`(connect timeout, read
                timeout) <timeouts>` tuple.
            :type timeout: float or tuple
            :param allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
            :type allow_redirects: bool
            :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
            :param verify: (optional) Either a boolean, in which case it controls whether we verify
                    the server's TLS certificate, or a string, in which case it must be a path
                    to a CA bundle to use. Defaults to ``True``.
            :param stream: (optional) if ``False``, the response content will be immediately downloaded.
            :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
            :param callback: callback routine used to handle errors
            :return: :class:`Response <Response>` object
            :rtype: requests.Response
        """
        try:
            if method.lower() not in ('get', 'post', 'put', 'patch', 'delete', 'head', 'options'):
                raise IntegrationError("unknown method {}".format(method))

            # If proxies was not set check if they are set in the config
            if proxies is None:
                proxies = self.get_proxies()

            if timeout is None:
                timeout = self.get_timeout()

            # Log the parameter inputs that are not None
            args_dict = locals()
            # When debugging execute_call_v2 in PyCharm you may get an exception when executing the for-loop:
            # Dictionary changed size during iteration.
            # To work around this while debugging you can change the following line to:
            # args = list(args_dict.keys())
            args = args_dict.keys()
            for k in args:
                if k != "self" and k != "kwargs" and args_dict[k] is not None:
                    log.debug("  {}: {}".format(k, args_dict[k]))

            # Pass request to requests.request() function
            response = requests.request(method, url, timeout=timeout, proxies=proxies, **kwargs)

            # Debug logging
            log.debug(response.status_code)
            log.debug(response.content)

            # custom handler for response handling
            # set callback to be the name of the method you would like to call
            # to do your custom error handling and return the response
            if callback:
                return callback(response)

            # Raise error is bad status code is returned
            response.raise_for_status()

            # Return requests.Response object
            return response

        except Exception as err:
            msg = str(err)
            log and log.error(msg)
            raise IntegrationError(msg)

    @deprecated("Use the new method execute_call_v2()")
    def execute_call(self, verb, url, payload={}, log=None, basicauth=None, verify_flag=True, headers=None,
                     proxies=None, timeout=None, resp_type='json', callback=None):
        """
        Function: perform the http API call. Different types of http operations are supported:
        GET, HEAD, PATCH, POST, PUT, DELETE
        Errors raise IntegrationError
        If a callback method is provided, then it's called to handle the error

        When using argument 'json' Content-Type will automatically be set to "application/json" by requests lib.

        :param verb: GET, HEAD, PATCH, POST, PUT, DELETE
        :param url:
        :param basicauth: used for basic authentication - (user, password)
        :param payload:
        :param log: optional log statement
        :param verify_flag: True/False - False used for debugging generally
        :param headers: dictionary of http headers
        :param proxies: http and https proxies for call
        :param timeout: timeout before call should abort
        :param resp_type: type of output to return: json, text, bytes
        :param callback: callback routine used to handle errors
        :return: json of returned data
        """

        try:
            (payload and log) and log.debug(payload)

            if verb.lower() not in ('get', 'post', 'put', 'patch', 'delete', 'head'):
                raise IntegrationError("unknown verb {}".format(verb))

            if proxies is None:
                proxies = self.get_proxies()

            if verb.lower() == 'post':

                content_type = get_case_insensitive_key_value(headers, "Content-Type")

                if is_payload_in_json(content_type):
                    resp = requests.request(verb.upper(), url, verify=verify_flag, headers=headers, json=payload,
                                            auth=basicauth, timeout=timeout, proxies=proxies)
                else:
                    resp = requests.request(verb.upper(), url, verify=verify_flag, headers=headers, data=payload,
                                            auth=basicauth, timeout=timeout, proxies=proxies)
            else:
                resp = requests.request(verb.upper(), url, verify=verify_flag, headers=headers, params=payload,
                                        auth=basicauth, timeout=timeout, proxies=proxies)

            if resp is None:
                raise IntegrationError('no response returned')

            # custom handler for response handling?
            if callback:
                return callback(resp)

            # standard error handling
            if resp.status_code >= 300:
                # get the result
                # log resp.status_code in case resp.text isn't available
                raise IntegrationError(
                    "status_code: {}, msg: {}".format(resp.status_code, resp.text if resp.text else "N/A"))

            # check if anything returned
            log and log.debug(resp.text)

            # get the result
            if resp_type == 'json':
                r = resp.json()
            elif resp_type == 'text':
                r = resp.text
            elif resp_type == 'bytes':
                r = resp.content
            else:
                raise IntegrationError("incorrect response type: {}".format(resp_type))

            # Produce a IntegrationError with the return value
            return r      # json object needed, not a string representation
        except Exception as err:
            msg = str(err)
            log and log.error(msg)
            raise IntegrationError(msg)


def is_payload_in_json(content_type):
    """
    Verify the content_type.
    If "Content-Type" is NOT specified pass the payload to the "json" argument - return True.

    If "Content-Type" is specified:
        - if the value is "application/json" pass the payload to the "json" argument - return True.
        - if the value is NOT "application/json" pass the payload to the "data" argument - return False.
    :param content_type:
    :return: True or False
    """

    if not content_type:
        return True

    return "application/json" in content_type.lower()


def get_case_insensitive_key_value(dictionary, key):
    """
    Get case insensitive key value from dictionary.
    :param dictionary:
    :param key:
    :return: value or None
    """
    if dictionary is None:
        return None

    return next((v for k, v in dictionary.items() if k.lower() == key.lower()), None)



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2018. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

import datetime
import tempfile
import os
import io
import mimetypes
import logging
import resilient
from bs4 import BeautifulSoup
from six import string_types
from cachetools import cached, TTLCache

try:
    from HTMLParser import HTMLParser as htmlparser
except:
    from html.parser import HTMLParser as htmlparser

INCIDENT_FRAGMENT = '#incidents'
PAYLOAD_VERSION = "1.0"

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(logging.StreamHandler())


def build_incident_url(url, incidentId):
    """
    build the url to link to an resilient incident
    :param url: base url
    :param incidentId:
    :return: full url
    """
    return '/'.join([url, INCIDENT_FRAGMENT, str(incidentId)])


def build_resilient_url(host, port):
    """
    build basic url to resilient instance
    :param host: host name
    :param port: port
    :return: base url
    """
    if host.lower().startswith("http"):
        return "{0}:{1}".format(host, port)

    return "https://{0}:{1}".format(host, port)


def clean_html(html_fragment):
    """
    Resilient textarea fields return html fragments. This routine will remove the html and insert any code within <div></div>
    with a linefeed
    :param html_fragment: str presenting the html to clean up
    :return: cleaned up code. This may not format well as no presentation of line feeds are preserved in the way supported by
       tags such as <br> or <ol>, <ul>, etc. See html2markdown for a better way to translate html input to markdown.
    """

    if not html_fragment or not isinstance(html_fragment, string_types):
        return html_fragment

    s = BeautifulSoup(unescape(html_fragment), "html.parser")

    return ' '.join(s.strings)


def unescape(data):
    """ Return unescaped data such as &gt; -> >, &quot -> ', etc.
    :param data: text to convert
    """
    if data is None:
        return None

    h = htmlparser()
    return h.unescape(data)


def validate_fields(field_list, kwargs):
    """
    Ensure each mandatory field in field_list is present in kwargs.
    Throw ValueError if not.

    field_list can be a list/tuple of strings where each string is
    a field name or it can be a list/tuple of dicts where each item
    has the attributes 'name' (required) and 'placeholder' (optional).

    If the value of the item in kwargs is equal to its placeholder
    defined in field_list, a ValueError is raised.

    If an item in kwargs is a Resilient Select Function Input, its
    value will be a dict that has a 'name' attribute. This returns
    the value of 'name'.

    If an item in kwargs is a Resilient Multi-Select Function Input, its
    value will be a list of dicts that have the 'name' attribute. This
    returns a list of the 'name' values for that item.

    :param field_list: list/tuple of the mandatory fields. Can be an empty list if no mandatory fields.
    :param kwargs: dict of all the fields to search.
    :return: a Dictionary of all fields with Select/Multi-Select fields handled.
    """

    mandatory_fields = field_list
    provided_fields = kwargs
    return_fields = {}
    mandatory_err_msg = "'{0}' is mandatory and is not set. You must set this value to run this function"

    # This is needed to handle something like: validate_fields(('incident_id'), kwargs)
    # In this case field_list will be a string and not a tuple
    if isinstance(mandatory_fields, string_types):
        mandatory_fields = [mandatory_fields]

    if not isinstance(mandatory_fields, list) and not isinstance(mandatory_fields, tuple):
        raise ValueError("'field_list' must be of type list/tuple, not {0}".format(type(mandatory_fields)))

    if not isinstance(provided_fields, dict):
        raise ValueError("'kwargs' must be of type dict, not {0}".format(type(provided_fields)))

    # Validate that mandatory fields exist + are not equal to their placeholder values
    for field in mandatory_fields:

        placeholder_value = None

        if isinstance(field, dict):
            placeholder_value = field.get("placeholder")
            field = field.get("name")

        # If the field value is a defined empty str, raise an error
        if isinstance(provided_fields.get(field), string_types):
            if not provided_fields.get(field):
                raise ValueError(mandatory_err_msg.format(field))

        if provided_fields.get(field) is None:
            raise ValueError(mandatory_err_msg.format(field))

        if placeholder_value and provided_fields.get(field) == placeholder_value:
            raise ValueError(
                "'{0}' is mandatory and still has its placeholder value of '{1}'. You must set this value correctly to run this function".format(
                    field, placeholder_value))

    # Loop provided fields and get their value
    for field_name, field_value in provided_fields.items():

        # Handle if Select Function Input type
        if isinstance(field_value, dict) and field_value.get("name"):
            field_value = field_value.get("name")

        # Handle if 'Text with value string Input' type
        elif isinstance(field_value, dict) and field_value.get("content"):
            field_value = field_value.get("content")

        # Handle if Multi-Select Function Input type
        elif isinstance(field_value, list):
            field_value = [f.get("name") for f in field_value]

        return_fields[field_name] = field_value

    return return_fields


def get_file_attachment(res_client, incident_id, artifact_id=None, task_id=None, attachment_id=None):
    """
    call the Resilient REST API to get the attachment or artifact data
    :param res_client: required for communication back to resilient
    :param incident_id: required
    :param artifact_id: optional
    :param task_id: optional
    :param attachment_id: optional
    :return: byte string of attachment
    """

    if incident_id and artifact_id:
        data_uri = "/incidents/{}/artifacts/{}/contents".format(incident_id, artifact_id)
    elif attachment_id:
        if task_id:
            data_uri = "/tasks/{}/attachments/{}/contents".format(task_id, attachment_id)
        elif incident_id:
            data_uri = "/incidents/{}/attachments/{}/contents".format(incident_id, attachment_id)
        else:
            raise ValueError("task_id or incident_id must be specified with attachment")
    else:
        raise ValueError("artifact or attachment or incident id must be specified")

    # Get the data
    return res_client.get_content(data_uri)


def get_file_attachment_metadata(res_client, incident_id, artifact_id=None, task_id=None, attachment_id=None):
    """
    call the Resilient REST API to get the attachment or artifact attachment metadata
    :param res_client: required for communication back to resilient
    :param incident_id: required
    :param artifact_id: optional
    :param task_id: optional
    :param attachment_id: optional
    :return: file attachment metadata
    """

    if incident_id and artifact_id:
        metadata_url = "/incidents/{}/artifacts/{}".format(incident_id, artifact_id)
        return res_client.get(metadata_url)["attachment"]

    if attachment_id:
        if task_id:
            metadata_url = "/tasks/{}/attachments/{}".format(task_id, attachment_id)
        elif incident_id:
            metadata_url = "/incidents/{}/attachments/{}".format(incident_id, attachment_id)
        else:
            raise ValueError("If attachment_id is defined, you must specify task_id OR incident_id")

        return res_client.get(metadata_url)

    raise ValueError("artifact_id AND incident_id, OR attachment_id AND (task_id OR incident_id) must be specified")


def get_file_attachment_name(res_client, incident_id=None, artifact_id=None, task_id=None, attachment_id=None):
    """
    call the Resilient REST API to get the attachment or artifact attachment name
    :param res_client: required for communication back to resilient
    :param incident_id: required
    :param artifact_id: optional
    :param task_id: optional
    :param attachment_id: optional
    :return: file attachment name
    """

    name = ""
    if incident_id and artifact_id:
        name_url = "/incidents/{}/artifacts/{}".format(incident_id, artifact_id)
        name = res_client.get(name_url)["attachment"]["name"]
    elif attachment_id:
        if task_id:
            name_url = "/tasks/{}/attachments/{}".format(task_id, attachment_id)
            name = res_client.get(name_url)["name"]
        elif incident_id:
            name_url = "/incidents/{}/attachments/{}".format(incident_id, attachment_id)
            name = res_client.get(name_url)["name"]
        else:
            raise ValueError("task_id or incident_id must be specified with attachment")
    else:
        raise ValueError("artifact or attachment or incident id must be specified")

    # Return name string
    return name


def write_file_attachment(res_client, file_name, datastream, incident_id, task_id=None, content_type=None):
    """
    call the Resilient REST API to create the attachment on incident or task

    :param res_client: required for communication back to resilient
    :param file_name: required, name of the attachment
    :param dataStream: required, stream of bytes
    :param incident_id: required
    :param task_id: optional
    :param content_type: optional, MIME type of attachment
    :return: new attachment -dictionary of attachment metadata
    """

    content_type = content_type \
                   or mimetypes.guess_type(file_name or "")[0] \
                   or "application/octet-stream"

    attachment = datastream.read()

    """
    Writing to temp path so that the REST API client can use this file path 
    to read and POST the attachment
    """

    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        try:
            temp_file.write(attachment)
            temp_file.close()

            # Create a new attachment by calling resilient REST API

            if task_id:
                attachment_uri = "/tasks/{}/attachments".format(task_id)
            else:
                attachment_uri = "/incidents/{}/attachments".format(incident_id)

            new_attachment = res_client.post_attachment(attachment_uri,
                                                        temp_file.name,
                                                        filename=file_name,
                                                        mimetype=content_type)
        finally:
            os.unlink(temp_file.name)

    if isinstance(new_attachment, list):
        new_attachment = new_attachment[0]

    return new_attachment


def readable_datetime(timestamp, milliseconds=True, rtn_format='%Y-%m-%dT%H:%M:%SZ'):
    """
    convert an epoch timestamp to a string using a format
    :param timestamp:
    :param milliseconds: True = epoch in
    :param rtn_format: format of resulant string
    :return: string representation of timestamp
    """
    if milliseconds:
        ts = int(timestamp / 1000)
    else:
        ts = timestamp

    return datetime.datetime.utcfromtimestamp(ts).strftime(rtn_format)


def str_to_bool(value):
    """Represents value as boolean.
    :param value:
    :rtype: bool
    """
    value = str(value).lower()
    return value in ('1', 'true', 'yes', 'on')


def write_to_tmp_file(data, tmp_file_name=None, path_tmp_dir=None):
    """Writes data to a file in a safely created temp directory. If no
    `tmp_file_name` is provided, a temp name will be given. If no `path_tmp_dir`
    is provided a temp directory is created with the prefix `resilient-lib-tmp-`.

    When used within a Resilient Function, ensure you safely remove the created temp
    directory in the `finally` block of the FunctionComponent code.

    Example:
        import os
        import shutil
        try:
            path_tmp_file, path_tmp_dir = write_to_tmp_file(attachment_contents, tmp_file_name=attachment_metadata.get("name"))

        except Exception:
            yield FunctionError()

        finally:
            if path_tmp_dir and os.path.isdir(path_tmp_dir):
                shutil.rmtree(path_tmp_dir)

    :param data: bytes to be written to the file
    :type data: `bytes`
    :param tmp_file_name: name to be given to the file.
    :type tmp_file_name: `str`
    :param path_tmp_dir: path to an existing directory to use as the temp dir
    :type path_tmp_dir: `str`
    :return: a tuple (path_tmp_file, path_tmp_dir)
    :rtype: tuple
    """

    # If no tmp_file_name provided use next tempfile candidate name
    if not tmp_file_name:
        tmp_file_name = next(tempfile._get_candidate_names())

    # If no path_tmp_dir provided, create one
    if not path_tmp_dir:
        path_tmp_dir = tempfile.mkdtemp(prefix="resilient-lib-tmp-")

    elif not os.path.isdir(path_tmp_dir):
        raise IOError("Path does not exist: {0}".format(path_tmp_dir))

    # Generate path to tmp file
    path_tmp_file = os.path.join(path_tmp_dir, tmp_file_name)

    # Write the file
    with io.open(path_tmp_file, mode="wb") as temp_file:
        temp_file.write(data)

    return (path_tmp_file, path_tmp_dir)


def close_incident(res_client, incident_id, kwargs):
    """
    :param res_client: required for communication back to resilient
    :param incident_id: required
    :param kwargs: required field_name:new_value pairs dict
    :return: response object
    """

    if not incident_id:
        raise ValueError("'incident_id' must be specified")

    # API call to the TypeRest for fields "required": "close" if not in kwargs throw an error
    required_fields = _get_required_fields(res_client)

    missing_fields = [field for field in required_fields if field not in kwargs]
    if missing_fields:
        raise ValueError("Missing mandatory field(s) to close an incident: {0}".format(missing_fields))

    # check for known mandatory field "plan_status" if not in kwargs add it
    mandatory_fields = kwargs.copy()
    if "plan_status" not in mandatory_fields:
        mandatory_fields["plan_status"] = "C"

    # API call to the Resilient REST API to patch the incident data (close incident)
    response = _patch_to_close_incident(res_client, incident_id, mandatory_fields)

    return response


def _get_required_fields(res_client):
    """
    :param res_client: required for communication back to resilient
    :return: list
    """
    fields = _get_incident_fields(res_client)
    fields_required = [field for field in fields if fields[field].get("required") == "close"]

    return fields_required


@cached(cache=TTLCache(maxsize=10, ttl=600))
def _get_incident_fields(res_client):
    """
    call the Resilient REST API to get list of fields required to close an incident
    this call is cached for multiple calls
    :param res_client: required for communication back to resilient
    :return: json
    """
    uri = "/types/incident"
    response = res_client.get(uri)
    incident_fields = response.get("fields")

    return incident_fields


def _patch_to_close_incident(res_client, incident_id, close_fields):
    """
    call the Resilient REST API to patch incident
    :param res_client: required for communication back to resilient
    :param incident_id: required
    :param close_fields: required
    :return: response object
    """
    uri = "/incidents/{}".format(incident_id)
    previous_object = res_client.get(uri)
    patch = resilient.Patch(previous_object)

    for field in close_fields:
        patch.add_value(field, close_fields[field])

    response = res_client.patch(uri, patch)

    return response



# -*- coding: utf-8 -*-

"""Generate a default configuration-file section for all integrations to use when proxies are involved"""


def config_section_data():
    """Produce the default configuration section for app.config,
       when called by `resilient-circuits config [-c|-u]`
    """
    config_data = u"""[integrations]
# These proxy settings will be used by all integrations. 
# To override, add any parameter to your specific integration section
http_proxy=
https_proxy=
timeout=30
"""
    return config_data


import unittest
from resilient_lib.components.function_metrics import FunctionMetrics


class TestFunctionMetrics(unittest.TestCase):
    """ Tests for the attachment_hash function"""

    def test_parameters(self):
        fm = FunctionMetrics("requests")
        result = fm.finish()

        self.assertIsNotNone(result)
        self.assertIsNotNone(result['package'])
        self.assertEqual(result['package'], 'requests')
        self.assertIsNotNone(result['package_version'])
        self.assertIsNotNone(result['host'])
        self.assertIsNotNone(result['execution_time_ms'])
        self.assertIsNotNone(result['timestamp'])

    def test_bad_pkg(self):
        fm = FunctionMetrics("missing")
        result = fm.finish()

        self.assertIsNotNone(result)
        self.assertIsNotNone(result['package'])
        self.assertEqual(result['package'], 'unknown')
        self.assertEqual(result['package_version'], 'unknown')
        self.assertIsNotNone(result['host'])
        self.assertIsNotNone(result['execution_time_ms'])
        self.assertIsNotNone(result['timestamp'])




# -*- coding: utf-8 -*-

import os
import sys
import shutil
import unittest
import logging
import pytest

from resilient_lib.components.resilient_common import str_to_bool, readable_datetime, validate_fields, \
    unescape, clean_html, build_incident_url, build_resilient_url, get_file_attachment, get_file_attachment_name, \
    get_file_attachment_metadata, write_to_tmp_file, close_incident

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(logging.StreamHandler())


class TestFunctionMetrics(unittest.TestCase):
    """ Tests for the attachment_hash function"""

    # List of temp directories to remove during tearDown
    DIRS_TO_REMOVE = []

    def setUp(self):
        # assertRaisesRegexp renamed to assertRaisesRegex in PY3.2
        if sys.version_info < (3, 2):
            self.assertRaisesRegex = self.assertRaisesRegexp

    @classmethod
    def tearDownClass(cls):
        # Clean up: remove any temporary directories created during tests
        for dirname in cls.DIRS_TO_REMOVE:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

    def test_str_to_bool(self):
        self.assertTrue(str_to_bool('True'))
        self.assertTrue(str_to_bool('true'))
        self.assertTrue(str_to_bool('YES'))
        self.assertFalse(str_to_bool('truex'))
        self.assertTrue(str_to_bool(1))
        self.assertTrue(str_to_bool('1'))
        self.assertFalse(str_to_bool(0))
        self.assertFalse(str_to_bool('0'))

    def test_readable_datetime(self):
        # readable_datetime(timestamp, milliseconds=True, str_format='%Y-%m-%dT%H:%M:%SZ'):

        ts = 1536194194
        str_ts = readable_datetime(ts, milliseconds=False)
        self.assertEqual(str_ts, '2018-09-06T00:36:34Z')

        str_ts = readable_datetime(ts * 1000, rtn_format='%Y-%m-%d')
        self.assertEqual(str_ts, '2018-09-06')

    def test_validate_fields(self):
        # validate_fields(fieldList, kwargs)

        inputs = {
            "bool_input_true": True,
            "bool_input_false": False,
            "unicode_input": u" դ ե զ է ը թ ժ ի լ խ ծ կ հ ձ ղ ճ մ յ ն ",
            "str_input": "some text",
            "num_input": 123,
            "select_input": {"id": 111, "name": "select choice"},
            "multi_select_input": [{"id": 111, "name": "select choice one"}, {"id": 111, "name": "select choice two"}],
            "text_with_value_string": {"content": "mock text", "format": "text"},
            "empty_input": ''
        }

        mandatory_fields = [
            {"name": "str_input", "placeholder": "some text"}
        ]

        expected_output = {
            "bool_input_true": True,
            "bool_input_false": False,
            "unicode_input": u" դ ե զ է ը թ ժ ի լ խ ծ կ հ ձ ղ ճ մ յ ն ",
            "str_input": "some text",
            "num_input": 123,
            "select_input": "select choice",
            "multi_select_input": ["select choice one", "select choice two"],
            "text_with_value_string": "mock text",
            "empty_input": ''
        }

        # Test its runs as expected
        validate_fields(("bool_input_true", "unicode_input"), inputs)

        with self.assertRaisesRegex(ValueError, "'field_list' must be of type list/tuple"):
            validate_fields({}, inputs)

        # Test mandatory fields missing
        with self.assertRaisesRegex(ValueError,
                                    "'cx' is mandatory and is not set. You must set this value to run this function"):
            validate_fields(("cx"), inputs)

        # Test mandatory field is empty string
        with self.assertRaisesRegex(ValueError,
                                    "'empty_input' is mandatory and is not set. You must set this value to run this function"):
            validate_fields(("empty_input"), inputs)

        # Test no mandatory fields
        self.assertEquals(validate_fields([], inputs), expected_output)

        # Test getting single input from returned dict
        self.assertEquals(validate_fields(("bool_input_true"), inputs).get("bool_input_true"), True)
        self.assertEquals(validate_fields([], inputs).get("bool_input_true"), True)

        # Test getting value defined as False
        self.assertEquals(validate_fields(["bool_input_false"], inputs).get("bool_input_false"), False)

        # Test select + multi-select type fields
        self.assertEquals(validate_fields(["select_input"], inputs).get("select_input"), "select choice")
        self.assertEquals(validate_fields([], inputs).get("multi_select_input"),
                          ["select choice one", "select choice two"])

        # Test 'Text with value string Input' type
        self.assertEquals(validate_fields(["text_with_value_string"], inputs).get("text_with_value_string"),
                          "mock text")

        # Test placeholder
        with self.assertRaisesRegex(ValueError,
                                    "'str_input' is mandatory and still has its placeholder value of 'some text'. You must set this value correctly to run this function"):
            validate_fields(mandatory_fields, inputs)

    def test_unescape(self):
        # unescape(data)
        test_data = "&lt;-&gt;"
        self.assertEqual(unescape(test_data), '<->')

        self.assertEqual(unescape("&ab;xx"), '&ab;xx')
        self.assertIsNone(unescape(None))

    def test_clean_html(self):
        # clean_html(htmlFragment)
        self.assertEqual(clean_html("<div>abc</div>"), "abc")
        self.assertEqual(clean_html("<div><ul><li>abc</li><li>def</li></ul></div>"), "abc def")
        self.assertEqual(clean_html("abc"), "abc")
        self.assertIsNone(clean_html(None))

    def test_build_incident_url(self):
        url = build_incident_url(build_resilient_url("https://localhost", 8443), 12345)
        self.assertEqual(url, "https://localhost:8443/#incidents/12345")

        url = build_incident_url(build_resilient_url("localhost", 8443), 12345)
        self.assertEqual(url, "https://localhost:8443/#incidents/12345")

    def test_file_attachment(self):
        with self.assertRaises(ValueError):
            get_file_attachment(None, 123)

        with self.assertRaises(ValueError):
            get_file_attachment(None, None, attachment_id=123)

    def test_file_attachment_name_error(self):
        with self.assertRaises(ValueError):
            get_file_attachment_name(None, 123)

        with self.assertRaises(ValueError):
            get_file_attachment_name(None, None, attachment_id=123)

    def test_get_file_attachment_metadata(self):
        with self.assertRaises(ValueError):
            get_file_attachment_metadata(res_client=None, incident_id=123)

        with self.assertRaises(ValueError):
            get_file_attachment_metadata(res_client=None, incident_id=None, attachment_id=123)

    def test_write_to_tmp_file(self):

        data_to_write = bytearray(u" դ ե զ է ը թ ժ ի լ խ ծ կ հ ձ ղ ճ մ յ ն ", encoding="utf-8")

        path_tmp_file, path_tmp_dir = write_to_tmp_file(data_to_write)

        self.DIRS_TO_REMOVE.append(path_tmp_dir)

        # Test works as expected
        self.assertTrue(os.path.isdir(path_tmp_dir))
        self.assertTrue(os.path.isfile(path_tmp_file))
        self.assertTrue("resilient-lib-tmp-" in path_tmp_file)

        # Test path does not exist
        with self.assertRaisesRegex(IOError, "Path does not exist:"):
            write_to_tmp_file(data_to_write, path_tmp_dir="xxxx")

    # Mocking attachment name tests using a dict
    def test_file_inc_attachment_name_str(self):
        inc_id = "1234"
        attachment_id = "123"
        expected_name = "This is a string attachment name"
        str_name_mock = {
            "/incidents/{}/attachments/{}".format(inc_id, attachment_id): {
                "name": expected_name
            }
        }
        actual_name = get_file_attachment_name(str_name_mock, inc_id, attachment_id=attachment_id)
        assert actual_name == expected_name

    def test_file_inc_attachment_name_unicode(self):
        inc_id = "1234"
        attachment_id = "123"
        expected_name = u'ÒÓı◊OIDÁÓØØÎ¨¨˝ Unicode attachment name'
        str_name_mock = {
            "/incidents/{}/attachments/{}".format(inc_id, attachment_id): {
                "name": expected_name
            }
        }
        actual_name = get_file_attachment_name(str_name_mock, inc_id, attachment_id=attachment_id)
        assert actual_name == expected_name

    def test_file_task_attachment_name_str(self):
        task_id = "1234"
        attachment_id = "123"
        expected_name = "This is a string attachment name"
        str_name_mock = {
            "/tasks/{}/attachments/{}".format(task_id, attachment_id): {
                "name": expected_name
            }
        }
        actual_name = get_file_attachment_name(str_name_mock, task_id=task_id, attachment_id=attachment_id)
        assert actual_name == expected_name

    def test_file_task_attachment_name_unicode(self):
        task_id = "1234"
        attachment_id = "123"
        expected_name = u'ÒÓı◊OIDÁÓØØÎ¨¨˝ Unicode attachment name'
        str_name_mock = {
            "/tasks/{}/attachments/{}".format(task_id, attachment_id): {
                "name": expected_name
            }
        }
        actual_name = get_file_attachment_name(str_name_mock, task_id=task_id, attachment_id=attachment_id)
        assert actual_name == expected_name

    def test_file_artifact_attachment_name_str(self):
        inc_id = "1234"
        artifact_id = "123"
        expected_name = "This is a string attachment name"
        str_name_mock = {
            "/incidents/{}/artifacts/{}".format(inc_id, artifact_id): {
                "attachment": {
                    "name": expected_name
                }
            }
        }
        actual_name = get_file_attachment_name(str_name_mock, incident_id=inc_id, artifact_id=artifact_id)
        assert actual_name == expected_name

    def test_file_artifact_attachment_name_unicode(self):
        inc_id = "1234"
        artifact_id = "123"
        expected_name = u'ÒÓı◊OIDÁÓØØÎ¨¨˝ Unicode attachment name'
        str_name_mock = {
            "/incidents/{}/artifacts/{}".format(inc_id, artifact_id): {
                "attachment": {
                    "name": expected_name
                }
            }
        }
        actual_name = get_file_attachment_name(str_name_mock, incident_id=inc_id, artifact_id=artifact_id)
        assert actual_name == expected_name

    @pytest.mark.skip(reason="fails on decorator")
    def test_close_incident(self):
        # patch_to_close_incident(res_client, incident_id, mandatory_fields):
        kwargs = {"resolution_summary": "<div class=\"rte\"><div>resolved</div></div>", "resolution_id": 2, "plan_status": "C"}
        incident_id = 123

        # Test incident_id missing
        with self.assertRaisesRegex(ValueError, "'incident_id' must be specified"):
            close_incident({}, None, kwargs)

        # Test mandatory fields missing
        kwargs_missing = {"resolution_summary": "<div class=\"rte\"><div>resolved</div></div>", "plan_status": "C"}
        mock_response = {
            "id": 0,
            "type_id": 0,
            "type_name": "incident",
            "fields": {
                "country": {"name": "country", "input_type": "select"},
                "resolution_id": {"name": "resolution_id", "input_type": "select", "required": "close"},
                "resolution_summary": {"name": "resolution_summary", "input_type": "textarea", "required": "close"},
                "workspace": {"name": "resolution_id", "input_type": "select", "required": "always"}
            },
            "resolution_summary": "<div class=\"rte\"><div>unresolved</div></div>",
            "vers": 5
        }
        mock_api = {
            "/types/incident": mock_response
        }
        with self.assertRaises(ValueError):
            close_incident(mock_api, incident_id, kwargs_missing)



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use
import resilient
import copy
from resilient_circuits.app import AppArgumentParser
from resilient_lib.components.resilient_common import str_to_bool
import logging

LOG = logging.getLogger(__name__)

TYPES_URL = "/types"
ORGANIZATION_TYPE = "organization"

LAYOUT_FOR_TYPE_QUERY = "/layouts?type={}"
LAYOUT_FOR = "/layouts/{}"


INCIDENT_TABS_NAME = "incident"
UI_LOCK = "ui_lock"


def get_organization_layout(client, layout_name):
    # first get a type ID for layout UI
    types = client.get(TYPES_URL)
    if ORGANIZATION_TYPE not in types:
        LOG.debug(types)
        raise KeyError("Can't find '{}' in the list of types.".format(ORGANIZATION_TYPE))

    organization = types[ORGANIZATION_TYPE]
    organization_type_id = organization["type_id"]
    # for organization extract UI layouts
    incident_layout = client.get(LAYOUT_FOR_TYPE_QUERY.format(organization_type_id))
    incident_tabs = [layout for layout in incident_layout if layout['name'] == layout_name]

    if not incident_tabs:
        raise ValueError("No UI tabs are received from the platform for {} layout.".format(layout_name))
    elif len(incident_tabs) != 1:
        raise ValueError("Expected 1 tab for {} layout, {} were returned".format(layout_name, len(incident_tabs)))

    return incident_tabs[0]


def get_incident_layout(client):
    return get_organization_layout(client, INCIDENT_TABS_NAME)


def get_incident_tabs(client):
    """
    Get a list of incident tabs exposed to user.
    :param client: Resilient client
    :returns: a tuple of layout id for incident UI and current list of UI tabs
    """
    tab_data = get_incident_layout(client)
    return tab_data.get('content')


def add_tab_to_layout(client, layout, new_tab):
    layout = copy.deepcopy(layout)
    layout['content'].append(
        new_tab
    )
    return client.put(LAYOUT_FOR.format(layout['id']), payload=layout)

def add_missing_fields(og_layout, tab):
    """
    Adds missing fields to the layout.
    """
    layout = copy.deepcopy(og_layout)
    missing_fields = tab.get_missing_fields(layout.get("content"))
    if not missing_fields:
        return og_layout
    tab_data = tab.get_from_tabs(layout.get("content"))
    tab_data['fields'].extend(missing_fields)
    return layout

def add_missing_conditions(og_layout, tab):
    """
    Adds missing conditions. Removes value = None from all existing condition, because that's what
    the UI does.
    """
    layout = copy.deepcopy(og_layout)
    tab_data = tab.get_from_tabs(layout.get("content"))
    # remove null values from conditions
    if tab_data.get('show_if'):
        for condition in tab_data.get('show_if'):
            if condition.get('value') is None:
                condition.pop('value')

    missing_conditions = tab.get_missing_conditions(layout.get("content"))
    if not missing_conditions:
        return layout  # still return updated one since we deleted nulls
    if not tab_data.get('show_if'):
        tab_data['show_if'] = []
    tab_data.get('show_if').extend(missing_conditions)
    return layout
    
def update_tab(client, layout, tab):
    """
    Needs to get the tab and add missing fields and conditions to it.
    """
    layout = copy.deepcopy(layout)
    layout = add_missing_fields(layout, tab)
    layout = add_missing_conditions(layout, tab)
    return client.put(LAYOUT_FOR.format(layout['id']), payload=layout)


def permission_to_edit(tab, opts):
    """
    Gets the config and determines if the UI had been locked or is allowed to be edited.
    :param tab: class or instance that's a subclass of ui.Tab
    :return: If app.config locks the tab or not
    """
    if opts.get(tab.SECTION):
        section_lock = str_to_bool(opts.get(tab.SECTION).get(UI_LOCK))
        return not section_lock
    if opts.get('integrations'):
        integrations_lock = str_to_bool(opts.get('integrations').get(UI_LOCK))
        return not integrations_lock
    if opts.get('resilient'):
        global_lock = str_to_bool(opts.get('resilient').get(UI_LOCK))
        return not global_lock

    return True


def create_tab(tab, update_existing=False):
    """
    If allowed by app.config - creates or updates a tab in the UI according to the
    specification passed in the class.
    Can be forbidden to make changes by adding `ui_lock=<true/on>` in app.config under integration section,
    resilient, or "integrations".

    :param tab: Subclass of ui.Tab that has required parameters and describes the required layout.
    :param update_existing: Defines the behavior if tab is already present in the system.
    Either simply leave it alone, or go through required elements and add those that are missing.
    """
    try:
        opts = _get_opts()
        if not permission_to_edit(tab, opts):
            LOG.info("No permission to edit UI for {}".format(tab.SECTION))
            return
        client = resilient.get_client(opts)
        layout = get_incident_layout(client)

        # check if tab already exists in the layout
        if tab.exists_in(layout.get("content")):
            if update_existing:
                LOG.info("UI tab for {} already exists. Checking for updates.".format(tab.SECTION))
                return update_tab(client, layout, tab)
            else:
                LOG.info("UI tab for {} already exists. Not updating.".format(tab.SECTION))
                return
        LOG.info("Creating a UI tab for {}".format(tab.SECTION))
        return add_tab_to_layout(client, layout, tab.as_dto())
    except Exception as e:
        LOG.error("Failed to create/update tab in the UI for {}".format(tab.SECTION))
        LOG.error(str(e))


def _get_opts():
    """
    Gets options from AppArgumentParser in the same manner as circuits does.
    """
    return AppArgumentParser().parse_args()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from os import path
import io

this_directory = path.abspath(path.dirname(__file__))


def gather_changes():
    filepath = './CHANGES'  # The file from which we will pull the changes
    with io.open(filepath) as fp:
        lines = fp.readlines()  # Take in all the lines as a list
        first_section = []
        for num, line in enumerate(lines, start=1):
            if "version" in lines[num] and num != 0:
                # Get all the lines from the start of the list until num-1.
                # This, along with the if statement above will ensure we only capture the most recent change.
                first_section = lines[:num-1]
                break
        # Return the section with a newline at the end
        return " \n ".join(first_section)


with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    readme_text = f.read()
    long_description = readme_text.replace('### Changelog', "### Recent Changes\n {}".format(gather_changes()))

setup(
    name='resilient_lib',
    use_scm_version={"root": "../", "relative_to": __file__},
    setup_requires=['setuptools_scm'],
    url='https://github.com/ibmresilient/resilient-circuits-packages',
    license='MIT',
    author='IBM Resilient',
    author_email='support@resilientsystems.com',
    description="library for resilient-circuits functions",
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'bs4',
        'resilient_circuits>=30.0.0',
        'six'
    ],
    packages=find_packages(),
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
    ],
    tests_require=['pytest>=3.0.0, <4.1.0'],
    entry_points={
        "resilient.lib.configsection": ["gen_config = resilient_lib.util.config:config_section_data"]
    }
)


# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2020. All Rights Reserved.
import time
import logging
import requests

log = logging.getLogger(__name__)


class OAuth2ClientCredentialsSession(requests.Session):
    """
    Wrapper around requests.Session that receives authentication tokens through
    'client_credentials' type of OAuth2 grant, and adds tokens to the requests made through the session,
    as well as attempts to keep track of the token and refresh it when the time comes.

    If proxies are defined, every request will use them.
    This session doesn't requests authorization from the user first, the scope should be pre-authorized.

    Usage:
    >>> api1 = OAuth2ClientCredentialsSession('https://example1.com/<tenant_id>/oauth/v2/',\
                        client_id='xxx', client_secret='xxx')
    >>> api2 = OAuth2ClientCredentialsSession('https://example2.com/<tenant_id>/oauth/v2/',\
                        client_id='xxx', client_secret='xxx')
    >>>
    >>> api1.post('https://example1.com/v4/me/messages', data={}) # use as a regular requests session object
    >>> api2.get('https://example2.com/v2/me/updates')
    >>> # When writing an integration, use RequestsCommon to get the proxies defined in in your app.config file.
    >>> rc = RequestsCommon(xxx)
    >>> api3 = OAuth2ClientCredentialsSession('https://example3.com/{}/test', proxies=rc.get_proxies())
    """

    AUTHORIZATION_ERROR_CODES = [401, 403]

    def __init__(self, url=None, client_id=None, client_secret=None, scope=None, proxies=None):
        """
        Get OAuth2 tokens and save them to be used in session requests.
        :param url: authorization url, with tenant_id in it, if required
        :param client_id: API key/User Id
        :param client_secret: secret for API
        :param scope: optional, list of scopes
        :param proxies: Dict - proxies to make the requests through
        """
        # __init__ is run even if __new__ returned existing instance
        if getattr(self, "authorization_url", None) is not None:
            return

        super(OAuth2ClientCredentialsSession, self).__init__()
        log.debug("Creating OAuth2 Session.")
        if not url or not client_id or not client_secret:
            missing_fields = list()
            if not url:
                missing_fields.append("url")
            if not client_id:
                missing_fields.append("client_id")
            if not client_secret:
                missing_fields.append("client_secret")
            raise ValueError("Missing fields ({}) required for OAuth2 authentication."
                             .format(",".join(missing_fields)))

        self.authorization_url = url
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope

        self.oauth_token_data = None
        self.expires_in = None
        self.access_token = None
        self.token_type = None
        self.expiration_time = None
        self.proxies = proxies
        self.close()  # release the socket, since all the requests will be made in a new session

        self.authenticate(url, client_id, client_secret, scope, proxies)

    def authenticate(self, url, client_id, client_secret, scope, proxies=None):
        """
        :param url: String - authorization url - end point for authentication
        :param client_id: String
        :param client_secret: String
        :param scope: list<String> - list of scopes the token should provide access to
        :param proxies: object with proxy data
        :return: True/False - was/wasn't able to authenticate.
        """
        token_url = url

        log.debug("Requesting token from {0}".format(url))
        r = self.get_token(token_url, client_id, client_secret, scope, proxies)

        log.info("Response status code: {}".format(r.status_code))
        r.raise_for_status()

        self.oauth_token_data = r.json()
        log.debug("Received oauth token.")

        self.expires_in = self.oauth_token_data.get("expires_in", None)
        if self.expires_in is not None:
            self.expiration_time = time.time() + int(self.expires_in)
        else:
            self.expiration_time = None

        self.access_token = self.oauth_token_data["access_token"]
        self.token_type = self.oauth_token_data.get("token_type", None)

        return True


    @staticmethod
    def get_token(token_url, client_id, client_secret, scope=None, proxies=None):
        """
        Override this method if there are any specific things that need to be changed.
        Such as Cloud Foundry asking grant_type to be 'password' and API key to be passed in 'password'.
        """
        post_data = {
            'client_id': client_id,
            'client_secret': client_secret,
            'grant_type': 'client_credentials'
        }
        if scope is not None:
            post_data['scope'] = scope
        return requests.post(token_url, data=post_data, proxies=proxies)

    def update_token(self):
        """
        Institutes a request for a new access token.
        """
        try:
            self.authenticate(self.authorization_url, self.client_id,
                                 self.client_secret, self.scope, self.proxies)
        except ValueError:
            raise ValueError("Can't update the token, did the credentials for {0} change?"
                             .format(self.authorization_url))
        return True

    def request(self, method, url, *args, **kwargs):
        """Constructs a :class:`Request <Request>`, injects it with tokens, sends it
        in a new session to avoid having one session constantly open.
        Returns :class:`Response <Response>` object.
        """

        if self.expiration_time is not None:
            if self.expiration_time < time.time():
                self.update_token()

        headers = kwargs.pop("headers") if "headers" in kwargs else {}

        self.add_authorization_header(headers)
        with requests.Session() as s:
            resp = s.request(method, url, *args, headers=headers, **kwargs)

        # If the error anything other than Authorization issue, the problem is in user's request
        if resp.status_code in self.AUTHORIZATION_ERROR_CODES:
            self.update_token()
            with requests.Session() as s:
                resp = s.request(method, url, *args, headers=headers, **kwargs)

        return resp

    def add_authorization_header(self, headers):
        """
        Create headers needed for authentication/authorization, overriding the default ones if needed.
        """
        headers["Authorization"] = "{} {}".format(self.token_type, self.access_token)
